@@ -191,6 +191,7 @@ gst_ffmpegaudenc_finalize (GObject * object)
   GstFFMpegAudEnc *ffmpegaudenc = (GstFFMpegAudEnc *) object;
 
   /* clean up remaining allocated data */
+  gst_ffmpeg_avcodec_close (ffmpegaudenc->context);
   av_free (ffmpegaudenc->context);
 
   G_OBJECT_CLASS (parent_class)->finalize (object);
@@ -203,6 +204,7 @@ gst_ffmpegaudenc_start (GstAudioEncoder * encoder)
   GstFFMpegAudEncClass *oclass =
       (GstFFMpegAudEncClass *) G_OBJECT_GET_CLASS (ffmpegaudenc);
 
+  gst_ffmpeg_avcodec_close (ffmpegaudenc->context);
   if (avcodec_get_context_defaults3 (ffmpegaudenc->context,
           oclass->in_plugin) < 0) {
     GST_DEBUG_OBJECT (ffmpegaudenc, "Failed to set context defaults");
@@ -237,15 +239,14 @@ gst_ffmpegaudenc_flush (GstAudioEncoder * encoder)
 static GstCaps *
 gst_ffmpegaudenc_getcaps (GstAudioEncoder * encoder, GstCaps * filter)
 {
-  GstFFMpegAudEnc *ffmpegaudenc = (GstFFMpegAudEnc *) encoder;
   GstCaps *caps = NULL;
 
-  GST_DEBUG_OBJECT (ffmpegaudenc, "getting caps");
+  GST_DEBUG_OBJECT (encoder, "getting caps");
 
   /* audio needs no special care */
   caps = gst_audio_encoder_proxy_getcaps (encoder, NULL, filter);
 
-  GST_DEBUG_OBJECT (ffmpegaudenc, "audio caps, return %" GST_PTR_FORMAT, caps);
+  GST_DEBUG_OBJECT (encoder, "audio caps, return %" GST_PTR_FORMAT, caps);
 
   return caps;
 }
@@ -297,7 +298,6 @@ gst_ffmpegaudenc_set_format (GstAudioEncoder * encoder, GstAudioInfo * info)
   ffmpegaudenc->context->coder_type = 0;
   ffmpegaudenc->context->context_model = 0;
   ffmpegaudenc->context->scenechange_threshold = 0;
-  ffmpegaudenc->context->inter_threshold = 0;
 
   /* fetch pix_fmt and so on */
   gst_ffmpeg_audioinfo_to_context (info, ffmpegaudenc->context);
@@ -406,6 +406,12 @@ gst_ffmpegaudenc_set_format (GstAudioEncoder * encoder, GstAudioInfo * info)
   return TRUE;
 }
 
+static void
+gst_ffmpegaudenc_free_avpacket (gpointer pkt)
+{
+  av_packet_unref ((AVPacket *) pkt);
+  g_slice_free (AVPacket, pkt);
+}
 
 static GstFlowReturn
 gst_ffmpegaudenc_encode_audio (GstFFMpegAudEnc * ffmpegaudenc,
@@ -416,7 +422,7 @@ gst_ffmpegaudenc_encode_audio (GstFFMpegAudEnc * ffmpegaudenc,
   gint res;
   GstFlowReturn ret;
   GstAudioInfo *info;
-  AVPacket pkt;
+  AVPacket *pkt;
   AVFrame frame;
   gboolean planar;
 
@@ -424,101 +430,112 @@ gst_ffmpegaudenc_encode_audio (GstFFMpegAudEnc * ffmpegaudenc,
 
   ctx = ffmpegaudenc->context;
 
-  GST_LOG_OBJECT (ffmpegaudenc, "encoding buffer ");
+  GST_LOG_OBJECT (ffmpegaudenc, "encoding buffer %p size:%u", audio_in,
+      in_size);
 
-  memset (&pkt, 0, sizeof (pkt));
-  memset (&frame, 0, sizeof (frame));
-  avcodec_get_frame_defaults (&frame);
+  pkt = g_slice_new0 (AVPacket);
 
-  info = gst_audio_encoder_get_audio_info (enc);
-  planar = av_sample_fmt_is_planar (ffmpegaudenc->context->sample_fmt);
+  if (audio_in != NULL) {
+    memset (&frame, 0, sizeof (frame));
+    avcodec_get_frame_defaults (&frame);
 
-  if (planar && info->channels > 1) {
-    gint channels, nsamples;
-    gint i, j;
+    info = gst_audio_encoder_get_audio_info (enc);
+    planar = av_sample_fmt_is_planar (ffmpegaudenc->context->sample_fmt);
 
-    nsamples = frame.nb_samples = in_size / info->bpf;
-    channels = info->channels;
+    if (planar && info->channels > 1) {
+      gint channels, nsamples;
+      gint i, j;
+
+      nsamples = frame.nb_samples = in_size / info->bpf;
+      channels = info->channels;
+
+      if (info->channels > AV_NUM_DATA_POINTERS) {
+        frame.extended_data = g_new (uint8_t *, info->channels);
+      } else {
+        frame.extended_data = frame.data;
+      }
+
+      frame.extended_data[0] = g_malloc (in_size);
+      frame.linesize[0] = in_size / channels;
+      for (i = 1; i < channels; i++)
+        frame.extended_data[i] = frame.extended_data[i - 1] + frame.linesize[0];
+
+      switch (info->finfo->width) {
+        case 8:{
+          const guint8 *idata = (const guint8 *) audio_in;
+
+          for (i = 0; i < nsamples; i++) {
+            for (j = 0; j < channels; j++) {
+              ((guint8 *) frame.extended_data[j])[i] = idata[j];
+            }
+            idata += channels;
+          }
+          break;
+        }
+        case 16:{
+          const guint16 *idata = (const guint16 *) audio_in;
+
+          for (i = 0; i < nsamples; i++) {
+            for (j = 0; j < channels; j++) {
+              ((guint16 *) frame.extended_data[j])[i] = idata[j];
+            }
+            idata += channels;
+          }
+          break;
+        }
+        case 32:{
+          const guint32 *idata = (const guint32 *) audio_in;
+
+          for (i = 0; i < nsamples; i++) {
+            for (j = 0; j < channels; j++) {
+              ((guint32 *) frame.extended_data[j])[i] = idata[j];
+            }
+            idata += channels;
+          }
+
+          break;
+        }
+        case 64:{
+          const guint64 *idata = (const guint64 *) audio_in;
+
+          for (i = 0; i < nsamples; i++) {
+            for (j = 0; j < channels; j++) {
+              ((guint64 *) frame.extended_data[j])[i] = idata[j];
+            }
+            idata += channels;
+          }
+
+          break;
+        }
+        default:
+          g_assert_not_reached ();
+          break;
+      }
 
-    if (info->channels > AV_NUM_DATA_POINTERS) {
-      frame.extended_data = g_new (uint8_t *, info->channels);
     } else {
+      frame.data[0] = audio_in;
       frame.extended_data = frame.data;
+      frame.linesize[0] = in_size;
+      frame.nb_samples = in_size / info->bpf;
     }
 
-    frame.extended_data[0] = g_malloc (in_size);
-    frame.linesize[0] = in_size / channels;
-    for (i = 1; i < channels; i++)
-      frame.extended_data[i] = frame.extended_data[i - 1] + frame.linesize[0];
+    /* we have a frame to feed the encoder */
+    res = avcodec_encode_audio2 (ctx, pkt, &frame, have_data);
 
-    switch (info->finfo->width) {
-      case 8:{
-        const guint8 *idata = (const guint8 *) audio_in;
-
-        for (i = 0; i < nsamples; i++) {
-          for (j = 0; j < channels; j++) {
-            ((guint8 *) frame.extended_data[j])[i] = idata[j];
-          }
-          idata += channels;
-        }
-        break;
-      }
-      case 16:{
-        const guint16 *idata = (const guint16 *) audio_in;
-
-        for (i = 0; i < nsamples; i++) {
-          for (j = 0; j < channels; j++) {
-            ((guint16 *) frame.extended_data[j])[i] = idata[j];
-          }
-          idata += channels;
-        }
-        break;
-      }
-      case 32:{
-        const guint32 *idata = (const guint32 *) audio_in;
-
-        for (i = 0; i < nsamples; i++) {
-          for (j = 0; j < channels; j++) {
-            ((guint32 *) frame.extended_data[j])[i] = idata[j];
-          }
-          idata += channels;
-        }
-
-        break;
-      }
-      case 64:{
-        const guint64 *idata = (const guint64 *) audio_in;
-
-        for (i = 0; i < nsamples; i++) {
-          for (j = 0; j < channels; j++) {
-            ((guint64 *) frame.extended_data[j])[i] = idata[j];
-          }
-          idata += channels;
-        }
-
-        break;
-      }
-      default:
-        g_assert_not_reached ();
-        break;
-    }
+    if (planar && info->channels > 1)
+      g_free (frame.data[0]);
+    if (frame.extended_data != frame.data)
+      g_free (frame.extended_data);
 
   } else {
-    frame.data[0] = audio_in;
-    frame.extended_data = frame.data;
-    frame.linesize[0] = in_size;
-    frame.nb_samples = in_size / info->bpf;
+    /* flushing the encoder */
+    res = avcodec_encode_audio2 (ctx, pkt, NULL, have_data);
   }
 
-  res = avcodec_encode_audio2 (ctx, &pkt, &frame, have_data);
-  if (planar && info->channels > 1)
-    g_free (frame.data[0]);
-  if (frame.extended_data != frame.data)
-    g_free (frame.extended_data);
-
   if (res < 0) {
     char error_str[128] = { 0, };
 
+    g_slice_free (AVPacket, pkt);
     av_strerror (res, error_str, sizeof (error_str));
     GST_ERROR_OBJECT (enc, "Failed to encode buffer: %d - %s", res, error_str);
     return GST_FLOW_OK;
@@ -529,20 +546,24 @@ gst_ffmpegaudenc_encode_audio (GstFFMpegAudEnc * ffmpegaudenc,
     GstBuffer *outbuf;
     const AVCodec *codec;
 
-    GST_LOG_OBJECT (ffmpegaudenc, "pushing size %d", pkt.size);
+    GST_LOG_OBJECT (ffmpegaudenc, "pushing size %d", pkt->size);
 
     outbuf =
-        gst_buffer_new_wrapped_full (0, pkt.data, pkt.size, 0, pkt.size,
-        pkt.data, av_free);
+        gst_buffer_new_wrapped_full (0, pkt->data, pkt->size, 0, pkt->size,
+        pkt, gst_ffmpegaudenc_free_avpacket);
 
     codec = ffmpegaudenc->context->codec;
-    if ((codec->capabilities & CODEC_CAP_VARIABLE_FRAME_SIZE)) {
+    if ((codec->capabilities & CODEC_CAP_VARIABLE_FRAME_SIZE) || !audio_in) {
+      /* FIXME: Not really correct, as -1 means "all the samples we got
+         given so far", which may not be true depending on the codec,
+         but we have no way to know AFAICT */
       ret = gst_audio_encoder_finish_frame (enc, outbuf, -1);
     } else {
       ret = gst_audio_encoder_finish_frame (enc, outbuf, frame.nb_samples);
     }
   } else {
     GST_LOG_OBJECT (ffmpegaudenc, "no output produced");
+    g_slice_free (AVPacket, pkt);
     ret = GST_FLOW_OK;
   }
 
